#ifndef TEST_VECTOR_HPP_
#define TEST_VECTOR_HPP_

#define CATCH_CONFIG_FAST_COMPILE
#define CATCH_CONFIG_DISABLE_MATCHERS
#define CATCH_CONFIG_CPP11_TO_STRING
#define CATCH_CONFIG_RUNNER
#include "catch.hpp"

#include <forward_list>
#include <initializer_list>
#include <vector>
#include <cstdlib>

#include "../Vector.hpp"

template <typename T>
struct Counter {
  T data;
  static int ctor, dtor, cctor, mctor, cassign, massign;
  Counter() { ++ctor; }
  Counter(Counter const &) { ++cctor; }
  Counter(Counter &&) { ++mctor; }
  Counter& operator=(Counter const&) { ++cassign; return *this; }
  Counter& operator=(Counter &&) { ++massign; return *this; }
  ~Counter() { ++dtor; }
  static void reset() {
    ctor = dtor = cctor = mctor = cassign = massign = 0;
  }
};

template <> int Counter<int>::ctor = 0;
template <> int Counter<int>::dtor = 0;
template <> int Counter<int>::cctor = 0;
template <> int Counter<int>::mctor = 0;
template <> int Counter<int>::cassign = 0;
template <> int Counter<int>::massign = 0;

class Range {
private:
  size_t low;
  size_t high;
public:
  Range () = delete;
  Range (size_t low, size_t high)
    : low (low), high (high) { }
  Range (size_t high)
    : low (0), high (high) { }
  size_t operator() () {
    return low + rand() % (high - low + 1);
  }
  static Range exactly (size_t n) {
    return Range (n, n);
  }
};

namespace wrap {

template <typename T>
struct Vector {
  using value_type = T;
  using size_type = std::size_t;
  using difference_type = std::ptrdiff_t;

  using reference = value_type&;
  using const_reference = value_type const&;

  using pointer = value_type*;
  using const_pointer = value_type const*;

  using iterator = pointer;
  using const_iterator = const_pointer;

  using reverse_iterator = std::reverse_iterator<iterator>;
  using const_reverse_iterator = std::reverse_iterator<const_iterator>;

  size_type m_size;
  size_type m_capacity;
  pointer m_data;
};

}

template <typename T>
wrap::Vector<T>&
access (Vector<T>& v) {
  return *reinterpret_cast<wrap::Vector<T>*> (&v);
}

template <typename T>
wrap::Vector<T> const&
access (Vector<T> const& v) {
  return *reinterpret_cast<wrap::Vector<T> const*> (&v);
}

template <typename T>
void
init (Vector<T>& v, Range&& size, Range&& capacity)
{
  size_t const SIZE = size();
  size_t const CAP = capacity();
  access(v).m_capacity = CAP;
  access(v).m_size = SIZE;
  delete[] std::exchange (access(v).m_data, new T[CAP]);
}

template <typename T, typename Fn>
void
fill (Vector<T>& v, Fn f)
{
  size_t const SIZE = access(v).m_size;
  T* const DATA = access(v).m_data;
  for (size_t i = 0; i < SIZE; ++i)
  {
    DATA[i] = f();
  }
}

SCENARIO ("Vector should be destructable", "[dtor]")
{
  Counter<int>::reset();
  GIVEN ("An empty Vector")
  {
    {
      Vector<Counter<int>> v;
    }
    WHEN ("The Vector is destructed")
    {
      THEN ("[1] The destructor isn't called for any elements")
      {
        REQUIRE (Counter<int>::dtor == 0);
      }
    }
  }
  GIVEN ("An non-empty Vector")
  {
    int const SIZE = 5;
    {
      Vector<Counter<int>> v (SIZE);
    }
    WHEN ("The Vector is destructed")
    {
      THEN ("[1] The destructor is called \"size\" times")
      {
        REQUIRE (Counter<int>::dtor == SIZE);
      }
    }
  }
}

SCENARIO ("Vector should have begin()", "[begin]")
{
  GIVEN ("A Vector with some size")
  {
    Vector<int> a (6);
    THEN ("[1] begin should point to the first element (if any)")
    {
      REQUIRE (a.begin() == access(a).m_data);
    }
  }
  GIVEN ("A const Vector with some size")
  {
    Vector<int> const a (7);
    THEN ("[1] begin should point to the first element (if any)")
    {
      REQUIRE (a.begin() == access(a).m_data);
    }
  }
}

SCENARIO ("Vector should have end()", "[end]")
{
  GIVEN ("A Vector with some size")
  {
    Vector<int> a (3);
    THEN ("[1] end should point to one-past-the-end")
    {
      REQUIRE (a.end() == (access(a).m_data + access(a).m_size));
    }
  }
  GIVEN ("A const Vector with some size")
  {
    Vector<int> const a (4);
    THEN ("[1] end should point to one-past-the-end")
    {
      REQUIRE (a.end() == (access(a).m_data + access(a).m_size));
    }
  }
}

SCENARIO ("Vector should have operator[]", "[access]")
{
  GIVEN ("A Vector with some size")
  {
    Vector<int> a(10);
    THEN ("[1] operator[] should return a correct reference")
    {
      REQUIRE (std::addressof(a[4]) == (access(a).m_data + 4));
    }
  }
  GIVEN ("A const Vector with some size")
  {
    Vector<int> const a(10);
    THEN ("[1] operator[] should return a correct reference")
    {
      REQUIRE (std::addressof(a[4]) == (access(a).m_data + 4));
    }
  }
}

SCENARIO ("Vector should have back()", "[back]")
{
  GIVEN ("A Vector with some size")
  {
    int const SIZE = 8;
    Vector<int> a(SIZE);
    THEN ("[1] back() should return a correct reference")
    {
      REQUIRE (std::addressof(a.back()) == (access(a).m_data + SIZE - 1));
    }
  }
  GIVEN ("A const Vector with some size")
  {
    int const SIZE = 10;
    Vector<int> const a(SIZE);
    THEN ("[1] back() should return a correct reference")
    {
      REQUIRE (std::addressof(a.back()) == (access(a).m_data + SIZE - 1));
    }
  }
}

SCENARIO ("Vector should have empty()", "[empty]")
{
  GIVEN ("An empty Vector")
  {
    Vector<int> a;
    THEN ("[1] empty() should return true")
    {
      REQUIRE (a.empty());
    }
  }
  GIVEN ("An non-empty Vector")
  {
    Vector<int> a(10);
    THEN ("[1] empty() should return false")
    {
      REQUIRE_FALSE (a.empty());
    }
  }
}

SCENARIO ("Vector should have capacity()", "[capacity]")
{
  GIVEN ("An empty Vector")
  {
    Vector<int> const a;
    THEN ("[1] capacity() should return the capacity of the Vector")
    {
      REQUIRE (a.capacity() == access(a).m_capacity);
    }
  }
  GIVEN ("A non-empty Vector")
  {
    Vector<int> const a(10);
    THEN ("[1] capacity() should return the capacity of the Vector")
    {
      REQUIRE (a.capacity() == access(a).m_capacity);
    }
  }
}

SCENARIO ("Vector should have clear()", "[clear]")
{
  GIVEN ("An empty Vector")
  {
    Vector<int> a;
    REQUIRE (a.size() == 0);
    REQUIRE (a.capacity() >= a.size());
    size_t oldCap = a.capacity();
    WHEN ("We clear() the vector")
    {
      a.clear();
      THEN ("[1] The size() should be 0 and capacity unchanged")
      {
        REQUIRE (a.size() == 0);
        REQUIRE (a.capacity() == oldCap);
      }
    }
  }
  GIVEN ("A Vector with some size")
  {
    int SIZE = rand() % 20 + 10;
    Vector<int> a(SIZE, 2);
    REQUIRE (a.size() == SIZE);
    REQUIRE (a.capacity() >= a.size());
    size_t oldCap = a.capacity();
    WHEN ("We clear() the vector")
    {
      a.clear();
      THEN ("[1] The size() should be 0 and capacity unchanged")
      {
        REQUIRE (a.size() == 0);
        REQUIRE (a.capacity() == oldCap);
      }
    }
  }
}

SCENARIO ("A Vector should have push_back(T const&)", "[push_back]")
{
  GIVEN ("An empty Vector")
  {
    size_t const SIZE = 0;
    Vector<int> a;
    REQUIRE (a.size() == SIZE);
    REQUIRE (a.capacity() >= SIZE);
    size_t const oldSize = a.size();
    size_t const oldCapacity = a.capacity();
    WHEN ("We push_back() a value")
    {
      int v = 4;
      a.push_back (v);
      THEN ("[1] The size() should increase by one and capacity should be 1")
      {
        REQUIRE (oldSize + 1 == a.size());
        REQUIRE (a.capacity() >= oldCapacity);
        REQUIRE (a.capacity() == 1);
        REQUIRE (access(a).m_data[SIZE] == v);
      }
    }
  }
  GIVEN ("A Vector with some size")
  {
    size_t const SIZE = rand() % 10 + 5;
    Vector<int> a (SIZE);;
    REQUIRE (a.size() == SIZE);
    REQUIRE (a.capacity() >= SIZE);
    size_t const oldSize = a.size();
    size_t const oldCapacity = a.capacity();
    WHEN ("We push_back() a value")
    {
      int v = 4;
      a.push_back (v);
      THEN ("[1] The size() should increase by one and capacity should double")
      {
        REQUIRE (oldSize + 1 == a.size());
        REQUIRE (a.capacity() == 2 * oldCapacity);
        REQUIRE (access(a).m_data[SIZE] == v);
      }
    }
  }
}

SCENARIO ("A Vector should have push_back(T&&)", "[push_back]")
{
  GIVEN ("An empty Vector")
  {
    size_t const SIZE = 0;
    Vector<int> a;
    REQUIRE (a.size() == SIZE);
    REQUIRE (a.capacity() >= SIZE);
    size_t const oldSize = a.size();
    size_t const oldCapacity = a.capacity();
    WHEN ("We push_back() a value")
    {
      int v = 4;
      int mv = v;
      a.push_back (std::move(mv));
      THEN ("[1] The size() should increase by one and capacity should be 1")
      {
        REQUIRE (oldSize + 1 == a.size());
        REQUIRE (a.capacity() >= oldCapacity);
        REQUIRE (a.capacity() == 1);
        REQUIRE (access(a).m_data[SIZE] == v);
      }
    }
  }
  GIVEN ("A Vector with some size")
  {
    size_t const SIZE = rand() % 10 + 5;
    Vector<int> a (SIZE);;
    REQUIRE (a.size() == SIZE);
    REQUIRE (a.capacity() >= SIZE);
    size_t const oldSize = a.size();
    size_t const oldCapacity = a.capacity();
    WHEN ("We push_back() a value")
    {
      int v = 4;
      int mv = v;
      a.push_back (std::move(mv));
      THEN ("[1] The size() should increase by one and capacity should double")
      {
        REQUIRE (oldSize + 1 == a.size());
        REQUIRE (a.capacity() == 2 * oldCapacity);
        REQUIRE (access(a).m_data[SIZE] == v);
      }
    }
  }
}

SCENARIO ("A Vector should have pop_back()", "[pop_back]")
{
  GIVEN ("A Vector with some elements")
  {
    Vector<int> a;
    init (a, Range::exactly(10), Range(10, 20));
    fill (a, [] { return rand() % 10; });
    size_t const old_cap = access(a).m_capacity;
    size_t const old_size = access(a).m_size;
    WHEN ("pop_back() is called")
    {
      a.pop_back();
      THEN ("[2] capacity is unchanged and size decreases")
      {
        REQUIRE (old_cap == a.capacity());
        REQUIRE (old_size - 1 == a.size());
      }
    }
  }
}

SCENARIO ("Vector should have a working copy constructor", "[copy-ctor]")
{
  GIVEN ("An empty Vector")
  {
    Vector<int> a;
    WHEN ("A new Vector is copy-constructed")
    {
      Vector<int> b(a);
      THEN ("[2] The two should be equivalent")
      {
        REQUIRE (a.size() == b.size());
        REQUIRE (b.capacity() >= b.size());
      }
    }
  }
  GIVEN ("A Vector with some elements")
  {
    Vector<int> a;
    init (a, Range(10, 20), Range(20, 30));
    fill (a, [] { return rand() % 10; });
    WHEN ("A new Vector is copy-constructed")
    {
      Vector<int> b(a);
      THEN ("[3] The two should be equivalent")
      {
        REQUIRE (a.size() == b.size());
        REQUIRE (b.capacity() >= b.size());
        size_t const elems = a.size();
        int* a_data = access(a).m_data;
        int* b_data = access(b).m_data;
        for (size_t i = 0; i < elems; ++i)
        {
          REQUIRE (a_data[i] == b_data[i]);
        }
      }
    }
  }
}


SCENARIO ("Vector should have a working range constructor", "[range-ctor]")
{
  GIVEN ("An empty Range")
  {
    std::vector<int> l;
    WHEN ("We range-construct the Vector")
    {
      Vector<int> v (l.begin(), l.end());
      THEN ("[2] the Vector should be empty")
      {
        REQUIRE (v.size() == 0);
        REQUIRE (v.capacity() >= v.size());
      }
    }
  }
  GIVEN ("A RandomAccessRange with elements")
  {
    std::vector<int> l = { 99, 88, 77, 66 };
    WHEN ("We range-construct the Vector")
    {
      Vector<int> v (l.begin(), l.end());
      THEN ("[3] the Vector should contain the same elements")
      {
        REQUIRE (v.size() == l.size());
        REQUIRE (v.capacity() >= v.size());
        REQUIRE (std::equal (v.begin(), v.end(), l.begin(), l.end()));
      }
    }
  }
  GIVEN ("A ForwardRange with elements")
  {
    std::forward_list<int> l = { 99, 88, 77, 66 };
    WHEN ("We range-construct the Vector")
    {
      Vector<int> v (l.begin(), l.end());
      THEN ("[2] the Vector should contain the same elements")
      {
        REQUIRE (v.size() == 4);
        REQUIRE (v.capacity() >= v.size());
        REQUIRE (std::equal (v.begin(), v.end(), l.begin(), l.end()));
      }
    }
  }
}

SCENARIO ("Vector should have a working copy assignment", "[copy-assign]")
{
  GIVEN ("Some Vector with elements")
  {
    Vector<int> a;
    init (a, Range(10,20), Range(20,30));
    size_t const CAP = access(a).m_capacity;
    size_t const SIZE = access(a).m_size;
    int* DATA = access(a).m_data;
    std::vector<int> v;
    fill (a, [&v, i=99] () mutable { v.push_back(i--); return v.back(); });
    WHEN ("We do self assignment")
    {
      auto& result = (v = v);
      THEN ("[2] Nothing should happen")
      {
        REQUIRE (access(a).m_size == SIZE);
        REQUIRE (access(a).m_capacity == CAP);
        REQUIRE (access(a).m_data == DATA);
        REQUIRE (std::equal (a.begin(), a.end(), v.begin(), v.end()));
      }
    }
    WHEN ("We do assignment from another Vector")
    {
      Vector<int> b;
      init (b, Range(100,200), Range(200,300));
      size_t const CAP2 = access(b).m_capacity;
      size_t const SIZE2 = access(b).m_size;
      int* DATA2 = access(b).m_data;
      fill (a, [i=1] () mutable { return i++; });

      a = b;

      THEN ("[3] The new vector should contain everything the old one did")
      {
        REQUIRE (access(a).m_size == access(b).m_size);
        REQUIRE (access(a).m_capacity >= access(a).m_size);
        REQUIRE (std::equal (a.begin(), a.end(), b.begin(), b.end()));
      }
    }
  }
}

SCENARIO ("Vector should have initializer_list assignment", "[init-list-assign]")
{
  GIVEN ("An empty Vector")
  {
    Vector<int> v;
    WHEN ("We assign it from an initializer_list { 1, 3, 5, 7}")
    {
      v = { 1, 3, 5, 7 };
      THEN ("[2] The contents should match")
      {
        REQUIRE (v.size() == 4);
        REQUIRE (v.capacity() >= v.size());
        REQUIRE (access(v).m_data[0] == 1);
        REQUIRE (access(v).m_data[1] == 3);
        REQUIRE (access(v).m_data[2] == 5);
        REQUIRE (access(v).m_data[3] == 7);
      }
    }
  }
  GIVEN ("A Vector with some elements")
  {
    Vector<int> v;
    init (v, Range(10, 20), Range(20, 40));
    fill (v, [] { return rand() % 10; });
    WHEN ("We assign it from an initializer_list { 1, 3, 5, 7}")
    {
      v = { 1, 3, 5, 7 };
      THEN ("[3] The contents should match")
      {
        REQUIRE (v.size() == 4);
        REQUIRE (v.capacity() >= v.size());
        REQUIRE (access(v).m_data[0] == 1);
        REQUIRE (access(v).m_data[1] == 3);
        REQUIRE (access(v).m_data[2] == 5);
        REQUIRE (access(v).m_data[3] == 7);
      }
    }
  }
}

SCENARIO ("Vector should implement shrink_to_fit", "[shrink_to_fit]")
{
  GIVEN ("A Vector with size = capacity")
  {
    Vector<int> a;
    init (a, Range::exactly(10), Range::exactly(10));
    fill (a, [] { return rand() % 10; });
    WHEN ("We call shrink_to_fit()")
    {
      size_t const SIZE = access(a).m_size;
      size_t const CAP = access(a).m_capacity;
      int* const DATA = access(a).m_data;
      a.shrink_to_fit();
      THEN ("[1] nothing should change")
      {
        REQUIRE (SIZE == access(a).m_size);
        REQUIRE (CAP == access(a).m_capacity);
        REQUIRE (DATA == access(a).m_data);
      }
    }
  }
  GIVEN ("A Vector with size < capacity")
  {
    Vector<int> a;
    init (a, Range(5, 10), Range(20, 30));

    std::vector<int> v;
    fill (a, [&v] {
               v.push_back (rand() % 10);
               return v.back();
             });

    WHEN ("We call shrink_to_fit()")
    {
      size_t const SIZE = access(a).m_size;
      size_t const CAP = access(a).m_capacity;
      int* const DATA = access(a).m_data;
      a.shrink_to_fit();
      THEN ("[4] data should be reallocated and copied; capacity should shrink")
      {
        REQUIRE (SIZE == access(a).m_size);
        REQUIRE (SIZE == access(a).m_capacity);
        REQUIRE (DATA != access(a).m_data);
        REQUIRE (std::equal (v.begin(), v.end(), a.begin(), a.end()));
      }
    }
  }
}

SCENARIO ("impl_safe_iterator() should have expected behaviors", "[iterator]")
{
  GIVEN ("A empty Vector (with size() == capacity() == 0)")
  {
    Vector<int> a;
    int* const DATA = access(a).m_data;
    size_t const SIZE = access(a).m_size;
    size_t const CAP = access(a).m_capacity;
    GIVEN ("n = 1")
    {
      size_t n = 1;
      WHEN ("impl_safe_iterator() is called")
      {
        auto iter = a.impl_safe_iterator(a.cbegin());
        THEN ("[2] data should have been reallocated")
        {
          REQUIRE (DATA != access(a).m_data);
          REQUIRE (SIZE == access(a).m_size);
          REQUIRE (CAP < access(a).m_capacity);
          REQUIRE (SIZE + n <= access(a).m_capacity);
          REQUIRE (iter == access(a).m_data);
        }
      }
    }
    GIVEN ("n > 1")
    {
      size_t n = 5;
      WHEN ("impl_safe_iterator() is called")
      {
        auto iter = a.impl_safe_iterator(a.cbegin(), n);
        THEN ("[2] data should have been reallocated, size unchanged, capacity changed")
        {
          REQUIRE (DATA != access(a).m_data);
          REQUIRE (SIZE == access(a).m_size);
          REQUIRE (CAP < access(a).m_capacity);
          REQUIRE (SIZE + n <= access(a).m_capacity);
          REQUIRE (iter == access(a).m_data);
        }
      }
    }
    GIVEN ("n >> 1")
    {
      size_t n = 1000;
      WHEN ("impl_safe_iterator() is called")
      {
        auto iter = a.impl_safe_iterator(a.cbegin(), n);
        THEN ("[2] data should have been reallocated, size unchanged, capacity changed")
        {
          REQUIRE (DATA != access(a).m_data);
          REQUIRE (SIZE == access(a).m_size);
          REQUIRE (CAP < access(a).m_capacity);
          REQUIRE (SIZE + n <= access(a).m_capacity);
          REQUIRE (iter == access(a).m_data);
        }
      }
    }
  }
  GIVEN ("A non-empty Vector with some size() + n <= capacity()")
  {
    Vector<int> a;
    init(a, Range::exactly(10), Range(20, 30));
    fill(a, [i = 0] () mutable {
              return ++i;
            });
    int* const DATA = access(a).m_data;
    size_t const SIZE = access(a).m_size;
    size_t const CAP = access(a).m_capacity;
    size_t const N = 15;
    size_t const INDEX = rand() % 5 + 5;
    WHEN ("impl_safe_iterator() is called")
    {
      auto iter = a.impl_safe_iterator(a.cbegin() + INDEX);
      THEN ("[1] Nothing should have changed")
      {
        REQUIRE (DATA == access(a).m_data);
        REQUIRE (SIZE == access(a).m_size);
        REQUIRE (CAP == access(a).m_capacity);
        REQUIRE (iter == access(a).m_data + INDEX);
        for (int i = 0; i < SIZE; ++i) {
          REQUIRE (access(a).m_data[i] == (i + 1));
        }
      }
    }
  }
  GIVEN ("size() + n > capacity()")
  {
    Vector<int> a;
    init(a, Range::exactly(10), Range(20, 30));
    fill(a, [i = 0] () mutable {
              return ++i;
            });
    int* const DATA = access(a).m_data;
    size_t const SIZE = access(a).m_size;
    size_t const CAP = access(a).m_capacity;
    size_t const N = 999;
    size_t const INDEX = rand() % 5 + 5;
    WHEN ("impl_safe_iterator() is called")
    {
      auto iter = a.impl_safe_iterator(a.cbegin() + INDEX, N);
      THEN ("[5] data should have been reallocated, size unchanged, capacity changed, elements copied")
      {
        REQUIRE (DATA != access(a).m_data);
        REQUIRE (SIZE == access(a).m_size);
        REQUIRE (CAP < access(a).m_capacity);
        REQUIRE (SIZE + N <= access(a).m_capacity);
        REQUIRE (iter == access(a).m_data + INDEX);
        for (int i = 0; i < SIZE; ++i) {
          REQUIRE (access(a).m_data[i] == (i + 1));
        }
      }
    }
  }
}

SCENARIO ("Vector should support erase() of single elements", "[erase-one]")
{
  GIVEN ("A Vector of some size")
  {
    Vector<int> a;
    size_t const CAP = 30;
    init (a, Range(10, 20), Range::exactly(CAP));
    std::vector<int> orig;
    fill (a, [&orig, i = 10] () mutable { orig.push_back(i++); return orig.back(); });
    // 10, 11, 12, ...
    WHEN ("We erase the first element")
    {
      auto iter = a.erase (a.begin());
      THEN ("[2] The first element was removed")
      {
        REQUIRE (access(a).m_data[0] == orig[1]);
        REQUIRE (access(a).m_size == orig.size() - 1);
        REQUIRE (access(a).m_capacity == CAP);
        REQUIRE (std::equal (a.begin(), a.end(), orig.begin() + 1, orig.end()));
        REQUIRE (a.begin() == iter);
      }
    }
    WHEN ("We erase the last element")
    {
      auto iter = a.erase (a.end() - 1);
      THEN ("[2] The last element was removed")
      {
        REQUIRE (access(a).m_data[access(a).m_size - 1] == *(orig.end() - 2));
        REQUIRE (access(a).m_size == orig.size() - 1);
        REQUIRE (access(a).m_capacity == CAP);
        REQUIRE (std::equal (a.begin(), a.end(), orig.begin(), orig.end() - 1));
        REQUIRE (a.end() == iter);
      }
    }
    WHEN ("We erase a middle element")
    {
      size_t const INDEX = 7;
      auto iter = a.erase (a.begin() + INDEX);
      THEN ("[2] A middle element was removed")
      {
        REQUIRE (access(a).m_data[INDEX] != orig[INDEX]);
        REQUIRE (access(a).m_size == orig.size() - 1);
        REQUIRE (access(a).m_capacity == CAP);
        REQUIRE (std::equal (a.begin(), a.begin() + INDEX, orig.begin(), orig.begin() + INDEX));
        REQUIRE (std::equal (a.begin() + INDEX, a.end(), orig.begin() + INDEX + 1, orig.end()));
        REQUIRE ((a.begin() + INDEX) == iter);
      }
    }
  }
}

SCENARIO ("Vector should support erase() of a Range", "[erase-range]")
{
  GIVEN ("A Vector of some size")
  {
    Vector<int> a;
    size_t const CAP = 10;
    init (a, Range(5, 10), Range::exactly(CAP));
    size_t const SIZE = access(a).m_size;

    std::vector<int> orig;
    fill (a, [&orig, i = 1] () mutable { orig.push_back(i++); return orig.back(); });
    // 1, 2, 3, ...

    WHEN ("We erase an empty range")
    {
      auto FIRST_ITER = a.begin() + 1;
      auto iter = a.erase (FIRST_ITER, FIRST_ITER);
      THEN ("[2] Nothing should change")
      {
        REQUIRE (SIZE == access(a).m_size);
        REQUIRE (CAP == access(a).m_capacity);
        REQUIRE (std::equal (a.begin(), a.end(), orig.begin(), orig.end()));
        REQUIRE (FIRST_ITER == iter);
      }
    }
    WHEN ("We erase a Range that goes to the end")
    {
      auto FIRST_ITER = a.begin() + 3;
      auto LAST_ITER = a.end();
      auto INDEX_OF_FIRST = std::distance (a.begin(), FIRST_ITER);
      auto INDEX_OF_LAST = std::distance (a.begin(), LAST_ITER);
      auto SIZE_OF_RANGE = std::distance (FIRST_ITER, LAST_ITER);
      auto iter = a.erase (FIRST_ITER, LAST_ITER);
      THEN ("[2] The last K elements should be erased")
      {
        REQUIRE (SIZE - SIZE_OF_RANGE == access(a).m_size);
        REQUIRE (CAP == access(a).m_capacity);
        REQUIRE (std::equal (a.begin(), iter, orig.begin(), orig.begin() + INDEX_OF_FIRST));
        REQUIRE (std::equal (iter, a.end(), orig.begin() + INDEX_OF_LAST, orig.end()));
        REQUIRE (FIRST_ITER == iter);
      }
    }
    WHEN ("We erase a Range at the beginning or middle")
    {
      auto FIRST_ITER = a.begin() + 1;
      auto LAST_ITER = a.begin() + 4;;
      auto INDEX_OF_FIRST = std::distance (a.begin(), FIRST_ITER);
      auto INDEX_OF_LAST = std::distance (a.begin(), LAST_ITER);
      auto SIZE_OF_RANGE = std::distance (FIRST_ITER, LAST_ITER);
      auto iter = a.erase (FIRST_ITER, LAST_ITER);
      THEN ("[2] The Range of elements should be removed and the rest moved down")
      {
        REQUIRE (SIZE - SIZE_OF_RANGE == access(a).m_size);
        REQUIRE (CAP == access(a).m_capacity);
        REQUIRE (std::equal (a.begin(), iter, orig.begin(), orig.begin() + INDEX_OF_FIRST));
        REQUIRE (std::equal (iter, a.end(), orig.begin() + INDEX_OF_LAST, orig.end()));
        REQUIRE (FIRST_ITER == iter);
      }
    }
    WHEN ("We erase a Range spanning the entire Vector")
    {
      auto iter = a.erase (a.begin(), a.end());
      THEN ("[2] The Vector should be empty")
      {
        REQUIRE (CAP == access(a).m_capacity);
        REQUIRE (0 == access(a).m_size);
        REQUIRE (iter == a.end());
      }
    }
  }
}

SCENARIO ("Vector should support resize() with default-constructed value", "[resize-default]")
{
  GIVEN ("A Vector of some size")
  {
    Vector<int> a;
    init(a, Range(7, 10), Range(15, 20));
    size_t const CAP = access(a).m_capacity;
    size_t const SIZE = access(a).m_size;
    int* const DATA = access(a).m_data;

    std::vector<int> b;
    fill(a, [&b, i = 0] () mutable {
              b.push_back(++i);
              return b.back();
            });

    WHEN ("We shrink the Vector")
    {
      a.resize(5);
      b.resize(5);
      THEN ("[1] size decreases and only elements past the new size are removed")
      {
        REQUIRE (a.size() == b.size());
        REQUIRE (a.capacity() == CAP);
        REQUIRE (access(a).m_data == DATA);
        REQUIRE (std::equal (a.begin(), a.end(), b.begin(), b.end()));
      }
    }

    WHEN ("We resize to the same size")
    {
      a.resize (a.size());
      THEN ("[1] nothing happens")
      {
        REQUIRE (a.size() == SIZE);
        REQUIRE (a.capacity() == CAP);
        REQUIRE (a.data() == DATA);
        REQUIRE (std::equal (a.begin(), a.end(), b.begin(), b.end()));
      }
    }

    WHEN ("We grow the vector but still have enough capacity")
    {
      a.resize (15);
      b.resize (15);
      THEN ("[1] we increase the size and add default-constructed elements to the end")
      {
        REQUIRE (a.size() == b.size());
        REQUIRE (a.capacity() == CAP);
        REQUIRE (access(a).m_data == DATA);
        REQUIRE (std::equal (a.begin(), a.end(), b.begin(), b.end()));
      }
    }

    WHEN ("We grow the vector but do not have enough capacity")
    {
      a.resize (30);
      b.resize (30);
      THEN ("[2] we reallocate, increase the size, and add default-constructed elements to the end")
      {
        REQUIRE (a.size() == b.size());
        REQUIRE (a.capacity() > CAP);
        REQUIRE (a.data() != DATA);
        REQUIRE (std::equal (a.begin(), a.end(), b.begin(), b.end()));
      }
    }
  }
}

SCENARIO ("Vector should support resize() with passed value", "[resize-value]")
{
  const int VALUE = 999;
  GIVEN ("A Vector of some size")
  {
    Vector<int> a;
    init(a, Range(7, 10), Range(15, 20));
    size_t const CAP = access(a).m_capacity;
    size_t const SIZE = access(a).m_size;
    int* const DATA = access(a).m_data;

    std::vector<int> b;
    fill(a, [&b, i = 0] () mutable {
              b.push_back(++i);
              return b.back();
            });

    WHEN ("We shrink the Vector")
    {
      a.resize(5, VALUE);
      b.resize(5, VALUE);
      THEN ("[1] size decreases and only elements past the new size are removed")
      {
        REQUIRE (a.size() == b.size());
        REQUIRE (a.capacity() == CAP);
        REQUIRE (access(a).m_data == DATA);
        REQUIRE (std::equal (a.begin(), a.end(), b.begin(), b.end()));
      }
    }

    WHEN ("We resize to the same size")
    {
      a.resize (a.size(), VALUE);
      THEN ("[1] nothing happens")
      {
        REQUIRE (a.size() == SIZE);
        REQUIRE (a.capacity() == CAP);
        REQUIRE (a.data() == DATA);
        REQUIRE (std::equal (a.begin(), a.end(), b.begin(), b.end()));
      }
    }

    WHEN ("We grow the vector but still have enough capacity")
    {
      a.resize (15, VALUE);
      b.resize (15, VALUE);
      THEN ("[1] we increase the size and add default-constructed elements to the end")
      {
        REQUIRE (a.size() == b.size());
        REQUIRE (a.capacity() == CAP);
        REQUIRE (access(a).m_data == DATA);
        REQUIRE (std::equal (a.begin(), a.end(), b.begin(), b.end()));
      }
    }

    WHEN ("We grow the vector but do not have enough capacity")
    {
      a.resize (30, VALUE);
      b.resize (30, VALUE);
      THEN ("[2] we reallocate, increase the size, and add default-constructed elements to the end")
      {
        REQUIRE (a.size() == b.size());
        REQUIRE (a.capacity() > CAP);
        REQUIRE (a.data() != DATA);
        REQUIRE (std::equal (a.begin(), a.end(), b.begin(), b.end()));
      }
    }
  }
}


#endif
